"""A library containing types of Spaces."""
import gym.spaces

try:
    import tensorflow as tf
except ImportError:  # pragma: no cover
    tf = False

try:
    import theano
except ImportError:  # pragma: no cover
    theano = False

from akro.box import Box
from akro.dict import Dict
from akro.discrete import Discrete
from akro.image import Image
from akro.space import Space
from akro.tuple import Tuple


def from_gym(space, is_image=False):
    """Convert a gym.space to an akro.space.

    Args:
        space(:obj:`gym.Space`): The Space object to convert.
        is_image(boolean): True if input is an image, False otherwise.
            This is False by default.

    Returns:
        akro.Space: The gym.Space object converted to an
            akro.Space object.

    """
    if isinstance(space, gym.spaces.Box):
        if is_image:
            assert (space.low == 0).all() and (space.high == 255).all(), \
                'Low and high values must be 0 and 255 for an Image'
            return Image(space.shape)
        return Box(low=space.low, high=space.high)
    elif isinstance(space, gym.spaces.Dict):
        return Dict(space.spaces)
    elif isinstance(space, gym.spaces.Discrete):
        return Discrete(space.n)
    elif isinstance(space, gym.spaces.Tuple):
        return Tuple(list(map(from_gym, space.spaces)))
    else:  # pragma: no cover
        raise TypeError


def concat(first, second):
    """Concatenate two spaces of the same type.

    Args:
        first (Space): The first space.
        second (Space): The second space.

    Returns:
        Space: A concatenated space.

    """
    assert type(first) == type(second)
    return first.concat(second)


__all__ = [
    'Space', 'Box', 'Dict', 'Discrete', 'Image', 'Tuple', 'from_gym', 'tf',
    'theano', 'concat'
]



import pickle
import unittest

import numpy as np

from akro import Image


class TestImage(unittest.TestCase):

    def test_pickleable(self):
        obj = Image((3, 3, 3))
        round_trip = pickle.loads(pickle.dumps(obj))
        assert round_trip
        assert round_trip.shape == obj.shape
        assert np.array_equal(round_trip.bounds[0], obj.bounds[0])
        assert np.array_equal(round_trip.bounds[1], obj.bounds[1])

    def test_invalid_shape(self):
        with self.assertRaises(AssertionError):
            Image((1, 1, 1, 1))
        with self.assertRaises(TypeError):
            Image()

    def test_dtype(self):
        img = Image((1, 1, 1))
        assert img.dtype == np.uint8

    def test_concat(self):
        img1 = Image((5, 5, 3))
        img2 = Image((10, 10, 3))
        concat_image = img1.concat(img2)

        assert concat_image.flat_dim == img1.flat_dim + img2.flat_dim



"""Cartesian product of multiple named Spaces (also known as a dict of Spaces).

This Space produces samples which are dicts, where the values of those dicts
are drawn from the values of this Space.
"""
import collections

import gym.spaces
import numpy as np

import akro
from akro.requires import requires_tf, requires_theano
from akro.space import Space


class Dict(gym.spaces.Dict, Space):
    """A dictionary of simpler spaces, e.g. Discrete, Box.

    Example usage:
        self.observation_space = spaces.Dict({"position": spaces.Discrete(2),
                                              "velocity": spaces.Discrete(3)})
    """

    def __init__(self, spaces=None, **kwargs):
        super().__init__(spaces, **kwargs)
        self.spaces = (collections.OrderedDict([
            (k, akro.from_gym(s)) for k, s in self.spaces.items()
        ]))

    @property
    def flat_dim(self):
        """Return the length of the flattened vector of the space."""
        return sum([space.flat_dim for _, space in self.spaces.items()])

    def flat_dim_with_keys(self, keys):
        """Return a flat dimension of the spaces specified by the keys.

        Returns:
            sum (int)

        """
        return sum([self.spaces[key].flat_dim for key in keys])

    def flatten(self, x):
        """Return an observation of x with collapsed values.

        Args:
            x (:obj:`Iterable`): The object to flatten.

        Returns:
            Dict: A Dict where each value is collapsed into a single dimension.
                  Keys are unchanged.

        """
        return np.concatenate(
            [space.flatten(x[key]) for key, space in self.spaces.items()],
            axis=-1,
        )

    def unflatten(self, x):
        """Return an unflattened observation x.

        Args:
            x (:obj:`Iterable`): The object to unflatten.

        Returns:
            collections.OrderedDict

        """
        dims = np.array([s.flat_dim for s in self.spaces.values()])
        flat_x = np.split(x, np.cumsum(dims)[:-1])
        return collections.OrderedDict([
            (key, self.spaces[key].unflatten(xi))
            for key, xi in zip(self.spaces.keys(), flat_x)
        ])

    def flatten_n(self, xs):
        """Return flattened observations xs.

        Args:
            xs (:obj:`Iterable`): The object to reshape and flatten

        Returns:
            np.ndarray: An array of xs in a shape inferred by the size of
                its first element.

        """
        return np.array([self.flatten(x) for x in xs])

    def unflatten_n(self, xs):
        """Return unflattened observations xs.

        Args:
            xs (:obj:`Iterable`): The object to reshape and unflatten

        Returns:
            List[OrderedDict]

        """
        return [self.unflatten(x) for x in xs]

    def flatten_with_keys(self, x, keys):
        """Return flattened obs of spaces specified by the keys using x.

        Returns:
            list

        """
        return np.concatenate(
            [
                space.flatten(x[key])
                for key, space in self.spaces.items() if key in keys
            ],
            axis=-1,
        )

    def unflatten_with_keys(self, x, keys):
        """Return an unflattened observation.

        This is the inverse of `flatten_with_keys`.

        Returns:
            collections.OrderedDict

        """
        dims = np.array([
            space.flat_dim for key, space in self.spaces.items() if key in keys
        ])
        flat_x = np.split(x, np.cumsum(dims)[:-1])
        return collections.OrderedDict([
            (key, space.unflatten(xi))
            for (key, space), xi in zip(self.spaces.items(), flat_x)
            if key in keys
        ])

    def concat(self, other):
        """Concatenate with another Dict space.

        If a key exists in both Dict, the two associated spaces will be concat.
        If a key exists in only one Dict, the associated space will be copied
        to the new Dict.

        Args:
            other (Dict): A space to be concatenated with this space.

        Returns:
            Dict: A concatenated space.

        """
        assert isinstance(other, Dict)

        spaces = dict()
        for key, space in self.spaces.items():
            if key in other.spaces:
                spaces[key] = space.concat(other.spaces[key])
            else:
                spaces[key] = space

        for key, space in other.spaces.items():
            if key not in self.spaces:
                spaces[key] = space

        return Dict(spaces)

    @requires_tf
    def to_tf_placeholder(self, name, batch_dims):
        """Create a tensor placeholder from the Space object.

        Args:
            name (str): name of the variable
            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of the object.

        Returns:
            tf.Tensor: Tensor object with the same properties as
                the Dict where the shape is modified by batch_dims.

        """
        newdict = Dict()
        for key, space in self.spaces.items():
            newdict.spaces[key] = space.to_tf_placeholder(name, batch_dims)
        return newdict

    @requires_theano
    def to_theano_tensor(self, name, batch_dims):
        """Create a theano tensor from the Space object.

        Args:
            name (str): name of the variable
            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of the object.

        Returns:
            theano.tensor.TensorVariable: Tensor object with the
                same properties as the Dict where the shape is
                modified by batch_dims.

        """
        newdict = Dict()
        for key, space in self.spaces.items():
            newdict.spaces[key] = space.to_theano_tensor(name, batch_dims)
        return newdict



import collections
import pickle
import unittest

import numpy as np

from akro import Box
from akro import Dict
from akro import Discrete
from akro import tf
from akro import theano
from akro.requires import requires_tf, requires_theano


class TestDict(unittest.TestCase):

    def test_pickleable(self):
        motion_dict = {'position': Discrete(2), 'velocity': Discrete(3)}
        sample = {
            'position': 1,
            'velocity': 2,
        }
        d = Dict(motion_dict)
        round_trip = pickle.loads(pickle.dumps(d))

        assert d.contains(sample)
        assert round_trip
        assert round_trip.contains(sample)

    def test_flat_dim(self):
        d = Dict(
            collections.OrderedDict(position=Box(0, 10, (2, )),
                                    velocity=Box(0, 10, (3, ))))
        assert d.flat_dim == 5

    def test_flat_dim_with_keys(self):
        d = Dict(
            collections.OrderedDict([('position', Box(0, 10, (2, ))),
                                     ('velocity', Box(0, 10, (3, )))]))
        assert d.flat_dim_with_keys(['position']) == 2

    def test_flatten(self):
        d = Dict(
            collections.OrderedDict([('position', Box(0, 10, (2, ))),
                                     ('velocity', Box(0, 10, (3, )))]))
        f = np.array([1., 2., 3., 4., 5.])
        # Keys are intentionally in the "wrong" order.
        s = collections.OrderedDict([('velocity', np.array([3., 4., 5.])),
                                     ('position', np.array([1., 2.]))])
        assert (d.flatten(s) == f).all()

    def test_unflatten(self):
        d = Dict(
            collections.OrderedDict([('position', Box(0, 10, (2, ))),
                                     ('velocity', Box(0, 10, (3, )))]))
        f = np.array([1., 2., 3., 4., 5.])
        # Keys are intentionally in the "wrong" order.
        s = collections.OrderedDict([('velocity', np.array([3., 4., 5.])),
                                     ('position', np.array([1., 2.]))])
        assert all((s[k] == v).all() for k, v in d.unflatten(f).items())

    def test_flatten_n(self):
        d = Dict(
            collections.OrderedDict([('position', Box(0, 10, (2, ))),
                                     ('velocity', Box(0, 10, (3, )))]))
        f = np.array([[1., 2., 3., 4., 5.], [6., 7., 8., 9., 0.]])
        # Keys are intentionally in the "wrong" order.
        s = [
            collections.OrderedDict([('velocity', np.array([3., 4., 5.])),
                                     ('position', np.array([1., 2.]))]),
            collections.OrderedDict([('velocity', np.array([8., 9., 0.])),
                                     ('position', np.array([6., 7.]))])
        ]
        assert (d.flatten_n(s) == f).all()

    def test_unflatten_n(self):
        d = Dict(
            collections.OrderedDict([('position', Box(0, 10, (2, ))),
                                     ('velocity', Box(0, 10, (3, )))]))
        f = np.array([[1., 2., 3., 4., 5.], [6., 7., 8., 9., 0.]])
        # Keys are intentionally in the "wrong" order.
        s = [
            collections.OrderedDict([('velocity', np.array([3., 4., 5.])),
                                     ('position', np.array([1., 2.]))]),
            collections.OrderedDict([('velocity', np.array([8., 9., 0.])),
                                     ('position', np.array([6., 7.]))])
        ]
        for i, fi in enumerate(d.unflatten_n(f)):
            assert all((s[i][k] == v).all() for k, v in fi.items())

    def test_flatten_with_keys(self):
        d = Dict(
            collections.OrderedDict([('position', Box(0, 10, (2, ))),
                                     ('velocity', Box(0, 10, (3, )))]))
        f = np.array([3., 4., 5.])
        f_full = np.array([1., 2., 3., 4., 5.])
        # Keys are intentionally in the "wrong" order.
        s = collections.OrderedDict([('velocity', np.array([3., 4., 5.])),
                                     ('position', np.array([1., 2.]))])
        assert (d.flatten_with_keys(s, ['velocity']) == f).all()
        assert (d.flatten_with_keys(s,
                                    ['velocity', 'position']) == f_full).all()

    def test_unflatten_with_keys(self):
        d = Dict(
            collections.OrderedDict([('position', Box(0, 10, (2, ))),
                                     ('velocity', Box(0, 10, (3, )))]))
        f = np.array([3., 4., 5.])
        f_full = np.array([1., 2., 3., 4., 5.])
        # Keys are intentionally in the "wrong" order.
        s = collections.OrderedDict([('velocity', np.array([3., 4., 5.])),
                                     ('position', np.array([1., 2.]))])
        assert all((s[k] == v).all()
                   for k, v in d.unflatten_with_keys(f, ['velocity']).items())
        assert all((s[k] == v).all() for k, v in d.unflatten_with_keys(
            f_full, ['velocity', 'position']).items())

    def test_concat(self):
        d1 = Dict(
            collections.OrderedDict([('position', Box(0, 10, (2, ))),
                                     ('velocity', Box(0, 10, (3, )))]))
        d2 = Dict(
            collections.OrderedDict([('position', Box(0, 10, (2, ))),
                                     ('gravity', Box(0, 10, (3, )))]))
        concat_d = d1.concat(d2)

        assert (sorted(concat_d.spaces.keys()) == sorted(
            ['position', 'velocity', 'gravity']))

    @requires_tf
    def test_convert_tf(self):
        d = Dict({'position': Discrete(2), 'velocity': Discrete(3)})
        tensor_dict = d.to_tf_placeholder('test', 1)
        assert isinstance(tensor_dict, Dict)
        assert all(
            [isinstance(c, tf.Tensor) for c in tensor_dict.spaces.values()])
        assert all([v.dtype == tf.int64 for v in tensor_dict.spaces.values()])

    @requires_theano
    def test_convert_theano(self):
        d = Dict({'position': Discrete(2), 'velocity': Discrete(3)})
        tensor_dict = d.to_theano_tensor('test', 1)
        assert isinstance(tensor_dict, Dict)
        assert all([
            isinstance(c, theano.tensor.TensorVariable)
            for c in tensor_dict.spaces.values()
        ])
        assert all(
            [space.dtype == 'int64' for space in tensor_dict.spaces.values()])



"""A space representing a selection between a finite number of items."""

import gym.spaces
import numpy as np

from akro import tf, theano
from akro.requires import requires_tf, requires_theano
from akro.space import Space


class Discrete(gym.spaces.Discrete, Space):
    """{0,1,...,n-1}."""

    def flatten(self, x):
        """Return a flattened observation x.

        Args:
            x (:obj:`Iterable`): The object to flatten.

        Returns:
            np.ndarray: An array of x collapsed into one dimension.

        """
        ret = np.zeros(self.n)
        ret[x] = 1
        return ret

    def unflatten(self, x):
        """Return an unflattened observation x.

        Args:
            x (:obj:`Iterable`): The object to unflatten.

        Returns:
            np.ndarray: An array of x in the shape of self.shape.

        """
        return np.nonzero(x)[0][0]

    def flatten_n(self, xs):
        """Return flattened observations xs.

        Args:
            xs (:obj:`Iterable`): The object to reshape and flatten

        Returns:
            np.ndarray: An array of xs in a shape inferred by the size of
                its first element.

        """
        ret = np.zeros((len(xs), self.n))
        ret[np.arange(len(xs)), xs] = 1
        return ret

    def unflatten_n(self, xs):
        """Return unflattened observations xs.

        Args:
            xs (:obj:`Iterable`): The object to reshape and unflatten

        Returns:
            np.ndarray: An array of xs in a shape inferred by the size of
                its first element and self.shape.

        """
        return np.nonzero(xs)[1]

    @property
    def flat_dim(self):
        """Return the length of the flattened vector of the space."""
        return self.n

    def weighted_sample(self, weights):
        """Compute a weighted sample of the elements in the Discrete Space.

        Args:
            weights (:obj:`list`): Values to use in the sample.

        Returns:
            int or np.ndarray: A random sample of n based on
                probabilities in weights.

        """
        assert len(weights) == self.n
        weights = np.asarray(weights)
        return np.random.choice(self.n, p=weights / weights.sum())

    def concat(self, other):
        """Concatenate with another space of the same type.

        Args:
            other (Space): A space to be concatenated with this space.

        Returns:
            Space: A concatenated space.

        """
        raise NotImplementedError

    def __hash__(self):
        """Hash the Discrete Space.

        Returns:
            int: A hash of the value n.

        """
        return hash(self.n)

    @requires_tf
    def to_tf_placeholder(self, name, batch_dims):
        """Create a tensor placeholder from the Space object.

        Args:
            name (str): name of the variable
            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of the object.

        Returns:
            tf.Tensor: Tensor object with the same properties as
                the Discrete obj where the shape is modified by batch_dims.

        """
        return tf.compat.v1.placeholder(dtype=self.dtype,
                                        shape=[None] * batch_dims +
                                        [self.flat_dim],
                                        name=name)

    @requires_theano
    def to_theano_tensor(self, name, batch_dims):
        """Create a theano tensor from the Space object.

        Args:
            name (str): name of the variable
            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of the object.

        Returns:
            theano.tensor.TensorVariable: Tensor object with the
                same properties as the Discrete obj where the shape is
                modified by batch_dims..

        """
        return theano.tensor.TensorType(self.dtype,
                                        (False, ) * (batch_dims + 1))(name)



import pickle
import unittest

import numpy as np

from akro import Box
from akro import tf
from akro import theano
from akro.requires import requires_tf, requires_theano


class TestBox(unittest.TestCase):

    def test_pickleable(self):
        obj = Box(-1.0, 1.0, (3, 4))
        round_trip = pickle.loads(pickle.dumps(obj))
        assert round_trip
        assert round_trip.shape == obj.shape
        assert np.array_equal(round_trip.bounds[0], obj.bounds[0])
        assert np.array_equal(round_trip.bounds[1], obj.bounds[1])

    def test_same_dtype(self):
        type1 = np.float32
        box = Box(0, 255, (3, 4), type1)
        assert box.dtype == type1

        type2 = np.uint8
        box = Box(0, 255, (3, 4), type2)
        assert box.dtype == type2

    def test_invalid_env(self):
        with self.assertRaises(AttributeError):
            Box(0.0, 1.0)

        with self.assertRaises(AssertionError):
            Box(np.array([-1.0, -2.0]), np.array([1.0, 2.0]), (2, 2))

    def test_default_float32_env(self):
        box = Box(0.0, 1.0, (3, 4))
        assert box.dtype == np.float32

        box = Box(np.array([-1.0, -2.0]), np.array([1.0, 2.0]))
        assert box.dtype == np.float32

    def test_flat_dim(self):
        box = Box(0.0, 1.0, (3, 4))
        assert box.flat_dim == 12

    def test_bounds(self):
        box = Box(0.0, 1.0, (3, 4))
        low, high = box.bounds
        assert low.shape == (3, 4)
        assert high.shape == (3, 4)
        assert low.dtype == np.float32
        assert high.dtype == np.float32

    def test_flatten(self):
        box = Box(0.0, 1.0, (3, 4))
        arr = [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]]
        flat_arr = box.flatten(arr)
        assert flat_arr.shape == (12, )
        assert flat_arr.dtype == np.int64

    def test_unflatten(self):
        box = Box(0.0, 1.0, (3, 4))
        arr = box.unflatten([(1, 2, 3), (4, 5, 6), (7, 8, 9), (10, 11, 12)])
        assert arr.shape == (3, 4)
        assert arr.dtype == np.int64

    def test_flatten_n(self):
        box = Box(0.0, 1.0, (3, 4))
        obs = np.asarray((([1, 2, 3], [3, 4, 5], [5, 6, 7], [7, 8, 9]),
                          ([1, 2, 3], [3, 4, 5], [5, 6, 7], [7, 8, 9]),
                          ([1, 2, 3], [3, 4, 5], [5, 6, 7], [7, 8, 9])))
        arr = box.flatten_n(obs)
        assert obs.shape == (3, 4, 3)
        assert arr.shape == (3, 12)

    def test_unflatten_n(self):
        box = Box(0.0, 1.0, (3, 4))
        obs = np.asarray((([1, 2, 3], [3, 4, 5], [5, 6, 7], [7, 8, 9]),
                          ([1, 2, 3], [3, 4, 5], [5, 6, 7], [7, 8, 9]),
                          ([1, 2, 3], [3, 4, 5], [5, 6, 7], [7, 8, 9])))
        arr = box.unflatten_n(obs)
        assert obs.shape == (3, 4, 3)
        assert arr.shape == (3, 3, 4)

    def test_flatten_n_tuple(self):
        box = Box(0.0, 1.0, (3, 4))
        obs = ((([1, 2, 3], [3, 4, 5], [5, 6, 7],
                 [7, 8, 9]), ([1, 2, 3], [3, 4, 5], [5, 6, 7], [7, 8, 9]),
                ([1, 2, 3], [3, 4, 5], [5, 6, 7], [7, 8, 9])))
        arr = box.flatten_n(obs)
        assert arr.shape == (3, 12)

    def test_unflatten_n_tuple(self):
        box = Box(0.0, 1.0, (3, 4))
        obs = ((([1, 2, 3], [3, 4, 5], [5, 6, 7],
                 [7, 8, 9]), ([1, 2, 3], [3, 4, 5], [5, 6, 7], [7, 8, 9]),
                ([1, 2, 3], [3, 4, 5], [5, 6, 7], [7, 8, 9])))
        arr = box.unflatten_n(obs)
        assert arr.shape == (3, 3, 4)

    def test_concat(self):
        box1 = Box(0.0, 1.0, (3, 4))
        box2 = Box(1.0, 2.0, (2, 3))
        concat_box = box1.concat(box2)

        expected_shape = box1.flat_dim + box2.flat_dim
        assert concat_box.shape == expected_shape
        assert np.array_equal(
            concat_box.low,
            np.concatenate([np.zeros(box1.flat_dim),
                            np.ones(box2.flat_dim)]))
        assert np.array_equal(
            concat_box.high,
            np.concatenate([np.ones(box1.flat_dim),
                            np.full(box2.flat_dim, 2)]))

    def test_hash(self):
        box = Box(0.0, 1.0, (3, 4))
        assert box.__hash__() == 1213972508617964782

    @requires_tf
    def test_convert_tf(self):
        box = Box(0.0, 1.0, (3, 4))
        tensor = box.to_tf_placeholder('test', 1)
        assert isinstance(tensor, tf.Tensor)
        assert tensor.dtype == tf.float32
        assert tensor.get_shape().as_list() == [None, 3, 4]

    @requires_theano
    def test_convert_theano(self):
        box = Box(0.0, 1.0, (3, 4))
        tensor = box.to_theano_tensor('test', 1)
        assert isinstance(tensor, theano.tensor.TensorVariable)
        assert box.dtype == np.float32
        assert tensor.dtype == 'float32'



"""Cartesian product of multiple Spaces (also known as a tuple of Spaces).

This Space produces samples which are Tuples, where the elments of those Tuples
are drawn from the components of this Space.
"""

import gym.spaces
import numpy as np

import akro
from akro.requires import requires_tf, requires_theano
from akro.space import Space


class Tuple(gym.spaces.Tuple, Space):
    """A Tuple of Spaces which produces samples which are Tuples of samples."""

    def __init__(self, spaces):
        super().__init__([akro.from_gym(space) for space in spaces])

    @property
    def flat_dim(self):
        """Return the length of the flattened vector of the space."""
        return np.sum([c.flat_dim for c in self.spaces])

    def flatten(self, x):
        """Return a flattened observation x.

        Args:
            x (:obj:`Iterable`): The object to flatten.

        Returns:
            np.ndarray: An array of x collapsed into one dimension.

        """
        return np.concatenate([c.flatten(xi) for c, xi in zip(self.spaces, x)])

    def flatten_n(self, obs):
        """Return flattened observations obs.

        Args:
            obs (:obj:`Iterable`): The object to reshape and flatten

        Returns:
            np.ndarray: An array of obs in a shape inferred by the size of
                its first element.

        """
        obs_regrouped = [[x[i] for x in obs] for i in range(len(obs[0]))]
        flat_regrouped = [
            c.flatten_n(xi) for c, xi in zip(self.spaces, obs_regrouped)
        ]
        return np.concatenate(flat_regrouped, axis=-1)

    def unflatten(self, x):
        """Return an unflattened observation x.

        Args:
            x (:obj:`Iterable`): The object to unflatten.

        Returns:
            tuple: A tuple of x in the shape of self.shape.

        """
        dims = [c.flat_dim for c in self.spaces]
        flat_x = np.split(x, np.cumsum(dims)[:-1])
        return tuple(c.unflatten(xi) for c, xi in zip(self.spaces, flat_x))

    def unflatten_n(self, obs):
        """Return unflattened observations obs.

        Args:
            obs (:obj:`Iterable`): The object to reshape and unflatten

        Returns:
            np.ndarray: An array of obs in a shape inferred by the size of
                its first element and self.shape.

        """
        dims = [c.flat_dim for c in self.spaces]
        flat_obs = np.split(obs, np.cumsum(dims)[:-1], axis=-1)
        unflat_obs = [
            c.unflatten_n(xi) for c, xi in zip(self.spaces, flat_obs)
        ]
        unflat_obs_grouped = list(zip(*unflat_obs))
        return unflat_obs_grouped

    def concat(self, other):
        """Concatenate with another Tuple space.

        Args:
            other (Tuple): A space to be concatenated with this space.

        Returns:
            Tuple: A concatenated space.

        """
        assert isinstance(other, Tuple)
        return Tuple(self.spaces + other.spaces)

    def __hash__(self):
        """Hash the Tuple Space.

        Returns:
            int: A hash of the Tuple's components.

        """
        return hash(tuple(self.spaces))

    @requires_tf
    def to_tf_placeholder(self, name, batch_dims):
        """Create a tensor placeholder from the Space object.

        Args:
            name (str): name to append to the akro type when naming
                the tensor.  e.g. When name is 'tmp' - 'Box-tmp',
                'Discrete-tmp'.

            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of each object in self.spaces.

        Returns:
            tuple(tf.Tensor): A tuple of Tensor objects converted
                from each Space in self.spaces. Each Tensor's
                shape is modified by batch_dims.

        """
        return tuple(
            s.to_tf_placeholder(type(s).__name__ + '-' + name, batch_dims)
            for s in self.spaces)

    @requires_theano
    def to_theano_tensor(self, name, batch_dims):
        """Create a theano tensor from the Space object.

        Args:
            name (str): name to append to the akro type when naming
                the tensor.  e.g. When name is 'tmp' - 'Box-tmp',
                'Discrete-tmp'.

            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of each object in self.spaces.

        Returns:
            theano.tensor.TensorVariable: A tuple of Tensor objects converted
                from each Space in self.spaces. Each Tensor's shape is
                modified by batch_dims.

        """
        return tuple(
            s.to_theano_tensor(type(s).__name__ + '-' + name, batch_dims)
            for s in self.spaces)



"""A Space representing an RGB Image."""
import numpy as np

from akro.box import Box


class Image(Box):
    """An Image, represented by a Box of at most three dimensions.

    This class allows us to type check the observation input
    and decide whether to normalize. Each dimension must have
    pixel values between [0, 255].

    Args:
        shape(tuple): Shape of the observation. The shape cannot
            have more than 3 dimensions.
    """

    def __init__(self, shape):
        assert len(shape) <= 3, 'Images must have at most three dimensions'
        super(Box, self).__init__(low=0, high=255, shape=shape, dtype=np.uint8)

    def concat(self, other):
        """Concatenate with another Image space.

        Args:
            other (Image): A space to be concatenated with this space.

        Returns:
            Image: A concatenated space.

        """
        assert isinstance(other, Image)
        return super().concat(other)



"""A Space representing a rectangular region of space."""
import gym.spaces
import numpy as np

from akro import tf, theano
from akro.requires import requires_tf, requires_theano
from akro.space import Space


class Box(gym.spaces.Box, Space):
    """A box in R^n.

    Each coordinate is bounded above and below.
    """

    @property
    def flat_dim(self):
        """Return the length of the flattened vector of the space."""
        return np.prod(self.low.shape)

    @property
    def bounds(self):
        """Return a 2-tuple containing the lower and upper bounds."""
        return self.low, self.high

    def flatten(self, x):
        """Return a flattened observation x.

        Args:
            x (:obj:'Iterable`): The object to flatten.

        Returns:
            np.ndarray: An array of x collapsed into one dimension.

        """
        return np.asarray(x).flatten()

    def unflatten(self, x):
        """Return an unflattened observation x.

        Args:
            x (:obj:`Iterable`): The object to unflatten.

        Returns:
            np.ndarray: An array of x in the shape of self.shape.

        """
        return np.asarray(x).reshape(self.shape)

    def flatten_n(self, obs):
        """Return flattened observations obs.

        Args:
            obs (:obj:`Iterable`): The object to reshape and flatten

        Returns:
            np.ndarray: An array of obs in a shape inferred by the size of
                its first element.

        """
        return np.asarray(obs).reshape((len(obs), -1))

    def unflatten_n(self, obs):
        """Return unflattened observation of obs.

        Args:
            obs (:obj:`Iterable`): The object to reshape and unflatten

        Returns:
            np.ndarray: An array of obs in a shape inferred by the size of
                its first element and self.shape.

        """
        return np.asarray(obs).reshape((len(obs), ) + self.shape)

    def concat(self, other):
        """Concatenate with another Box space.

        Note that the dimension of both boxes will be flatten.

        Args:
            other (Box): A space to be concatenated with this space.

        Returns:
            Box: A concatenated space.

        """
        assert isinstance(other, Box)

        first_lb, first_ub = self.bounds
        second_lb, second_ub = other.bounds
        first_lb, first_ub = first_lb.flatten(), first_ub.flatten()
        second_lb, second_ub = second_lb.flatten(), second_ub.flatten()
        return Box(np.concatenate([first_lb, second_lb]),
                   np.concatenate([first_ub, second_ub]))

    def __hash__(self):
        """Hash the Box Space.

        Returns:
            int: A hash of the low, high, and shape of the Box.

        Only the first element of low and high are hashed because numpy
        ndarrays can't be hashed. When a Box is created the low and high
        bounds are duplicated across the shape of the arrays so any of the
        values will suffice for the hash. The shape of the Box is added
        for uniqueness.

        """
        return hash((self.low[0][0], self.high[0][0], self.shape))

    @requires_tf
    def to_tf_placeholder(self, name, batch_dims):
        """Create a tensor placeholder from the Space object.

        Args:
            name (str): name of the variable
            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of the object.

        Returns:
            tf.Tensor: Tensor object with the same properties as
                the Box where the shape is modified by batch_dims.

        """
        return tf.compat.v1.placeholder(dtype=self.dtype,
                                        shape=[None] * batch_dims +
                                        list(self.shape),
                                        name=name)

    @requires_theano
    def to_theano_tensor(self, name, batch_dims):
        """Create a theano tensor from the Space object.

        Args:
            name (str): name of the variable
            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of the object.

        Returns:
            theano.tensor.TensorVariable: Tensor object with the
                same properties as the Box where the shape is
                modified by batch_dims.

        """
        return theano.tensor.TensorType(self.dtype,
                                        (False, ) * (batch_dims + 1))(name)
